#!/usr/bin/ruby

require 'strscan'

# This class is responsible for making the mediainfo from XMMS2 look neat & tidy
# on the playlist, mainwin, and wherever else. A user-supplied string determines
# exactly how each playlist entry should look. By default, the string is
# '%?(%t){%?(%a){%a - }{}%t}{%f}%?(%c){ [%c]}{}'. It might be self-
# explanatory to some, but I'll give an overview of how it works here. This
# string, called the display string format, contains %-escaped commands that
# allow conditionals and substitutions. Everything else in the display string is
# taken literally. %?(){}{} is the basic conditional construct. It takes a
# string for every argument, like so: '%?(<eval>){<true>}{<false>}'. The first
# argument <eval> is what the parser evaluates to determine which of the next
# two arguments to substitute (<true> or <false>). Any string in the parenthesis
# that does not evaluate to an empty string ('') results in the substitution of
# the string within the first set of curly braces (<true>). Thus, it's not wise
# to use any literal strings in the parenthesis and only %-escaped commands. For
# example, '%?(blah){foo}{bar}' will always evaluate to 'foo'. If <eval>
# evaluates to an empty string, then the second set of curly braces is
# substituted (<false>). It's basically a simplified, serialized if..then..else
# construct. Nesting is allowed and the characters %, (, ), {, and } can be used
# literally if they are escaped with a %.
#
# The other %-commands in the display string are substitutions. You can see
# the meaning of each one in the prettify method.

class Format
	def initialize(str)
		@str = str
		ary = []
		temp = ''
		scan = StringScanner.new(@str)

		# This parser scans the display string format for special %-escaped
		# characters and the () and {} parts of the conditional construct. The
		# whole string is then broken up into an array of literal strings, %-
		# escaped characters, and conditional construct pieces. This will aid
		# the parser later when it is converted to ruby code.
		loop do
			scan.scan(/(?:\%.)|\(|\)|\{|\}/)
			if(scan.matched?)
				ary.push temp unless(temp == '') # add any literal text first
				ary.push scan.matched
				temp = ''
			else
				temp += scan.peek(1) # if it didn't match, it must be literal
			end
			break if(scan.eos?)
			scan.pos = scan.pos + 1 unless(scan.matched?)
		end
		ary.push temp unless(temp == '')

		# That was probably hard to understand, so let me explain what just went
		# on. Basically, we have a string coming in that looks like this:
		# 'foo %?(bar){baz}{quux} fubar'
		# this loop just parsed it into an array like this:
		# ['foo ', '%?(', 'bar', ')', '{', 'baz', '}{', 'quux', '}', ' fubar']
		# This is going to help us a lot in the following code that actually
		# converts that into ruby code that can be evaluated directly. You may
		# think that generating ruby code within a ruby application is a rather
		# inelegant solution here, and you're right. But, more importantly,
		# since it's ruby code being evaluated and it doesn't use any slow,
		# memory-hungry regexes, it's very fast and lean. There are no
		# substitutions (in the String#sub sense), only conditionals and
		# concatenations. And this code generation is a one-time-only deal.

		@eval_str = ''
		ary.each_with_index do |str, i|
			case(str)
				when('%?')
					@eval_str.concat("if") if(ary[i + 1] == '(')
				when('(')
					@eval_str.concat("(") if(i > 0 && ary[i - 1] == '%?')
				when(')')
					@eval_str.concat(" != '')\n")
				when('{')
					@eval_str.concat("''") if(ary[i + 1] == '}')
				when('}')
					if(ary[i + 1] == '{')
						@eval_str.concat("\nelse\n")
					elsif(ary[i + 1])
						@eval_str.concat("\nend +\n")
					else
						@eval_str.concat("\nend\n")
					end
				when(/\%./)
					ret = "@substitutions['#{str}']"
					ret = ret + ' + ' if(ary[i + 1] &&
						ary[i + 1].match(/\)|\}/).nil?)
					@eval_str.concat(ret)
				else
					ret = "'" + str + "'"
					ret = ret + ' + ' if(ary[i + 1] &&
						ary[i + 1].match(/\)|\}/).nil?)
					@eval_str.concat(ret)
			end
		end
		# Uncomment this to see the ruby code that has been generated.
		# puts @eval_str
	end

		# The above isn't worth explaining bit-by-bit, the code is easy enough
		# to read if you really want to understand. However, the basic idea is
		# to take our array:
		# ['foo ', '%?(', 'bar', ')', '{', 'baz', '}{', 'quux', '}', ' fubar']
		# and transformed it into ruby code like the following:
		# 'foo ' +
		# if('bar' != '')
		# 'baz'
		# else
		# 'quux'
		# end +
		# ' fubar'
		# Which, when evaluated, will always print 'foo baz fubar'. (This isn't
		# very useful as anything but an example.) A real display string would
		# contain some %-values to be substituted. For example, %a is the
		# artist. The ruby code this would produce looks something like this:
		# @substitutions['%a']
		# Which would produce the artist for that particular entry.

	def prettify(hash)
		@substitutions = {
			'%a' => hash[:artist] || '',
			'%b' => hash[:bitrate] ? hash[:bitrate].to_s : '',
			'%c' => hash[:channel] || '',
			'%d' => hash[:date] ? hash[:date].to_s : '',
			'%f' => get_filename(hash[:url_decoded] || ''),
			'%b' => hash[:id] ? hash[:id].to_s : '',
			'%l' => hash[:album] || '',
			'%n' => hash[:tracknr] ? hash[:tracknr].to_s : '',
			'%p' => get_path(hash[:url_decoded] || ''),
			'%t' => hash[:title] || '',
			'%u' => hash[:url] || '',
			'%x' => get_ext(hash[:url_decoded] || ''),
			'%%' => '%',
			'%{' => '{',
			'%}' => '}',
			'%(' => '(',
			'%)' => ')'
		} # the substitutions are set once for every display string

		begin
			str = eval(@eval_str)
		rescue SyntaxError
			return nil
		end

		return str
	end

	private

	# This gets the path part of a file. File#dirname sucks, so we substitute.
	def get_path(url)
		return url.sub(/#{Regexp::escape(File.basename(url))}$/, '')
	end

	# This strips the path and extention from a filename.
	def get_filename(url)
		return File.basename(url).sub(/#{Regexp::escape(get_ext(url))}$/, '')
	end

	# And this returns the extention from a filename.
	def get_ext(url)
		return File.extname(url)
	end
end
