#!/usr/bin/env ruby

require 'xmmsclient'
require 'xmmsclient_glib'

class Xmms2
	attr_reader :client, :playlist, :pos, :id, :length
	# For the playlist callback
	ADD = XmmsClient::Result::PLAYLIST_CHANGED_ADD
	INS = XmmsClient::Result::PLAYLIST_CHANGED_INSERT
	MOV = XmmsClient::Result::PLAYLIST_CHANGED_MOVE
	REM = XmmsClient::Result::PLAYLIST_CHANGED_REMOVE
	CLR = XmmsClient::Result::PLAYLIST_CHANGED_CLEAR
	SRT = XmmsClient::Result::PLAYLIST_CHANGED_SORT
	SHF = XmmsClient::Result::PLAYLIST_CHANGED_SHUFFLE
	# For the status callback
	PLAYING = XmmsClient::XmmsClient::PLAY
	PAUSED = XmmsClient::XmmsClient::PAUSE
	STOPPED = XmmsClient::XmmsClient::STOP
	# Used by all callbacks
	UNKNOWN = -1

	def initialize(client, ipcpath)
		@client, @ipcpath = client, ipcpath

		init_callbacks
		try_connect
	end

	def medialib_add_key(id, key, data)
		# Mediainfo from the daemon is under the :server key. All other
		# mediainfo is stored under :client/<clientname>.
		if(reconnect_if_disconnected)
			@xc.medialib_entry_property_set(id, key, data)

			return true
		else
			return false
		end
	end

	def register_callback(key, &cb)
		return nil unless(@callbacks.has_key?(key))
		@callbacks[key].push(cb)	# callbacks are registered in order

		return cb.to_s
	end

	def unregister_callback(key, id)
		return false unless(@callbacks.has_key?(key))
		# To delete a specific callback and not every callback for a particular
		# broadcast, a copy of the callback to be deleted is required.
		return false unless(@callbacks[key].delete_if {|cb| cb.to_s == id})

		return true
	end

	# <convenience methods>
	def connected?
		return @connected
	end

	def playing?
		return @status == PLAYING
	end

	def paused?
		return @status == PAUSED
	end

	def stopped?
		return @status == STOPPED
	end

	def toggle_pause
		reconnect_if_disconnected if(@status == UNKNOWN)
		if(playing?)
			pause
		elsif(paused?)
			play
		end
	end
	# </convenience methods>

	def prev
		jump_rel(-1)	# a little more convenience
	end

	def play
		if(reconnect_if_disconnected)
			@xc.playback_start

			return true
		else
			return false
		end
	end

	def pause
		if(reconnect_if_disconnected)
			@xc.playback_pause

			return true
		else
			return false
		end
	end

	def stop
		if(reconnect_if_disconnected)
			@xc.playback_stop

			return true
		else
			return false
		end
	end

	def next
		jump_rel(1)		# aaaand a little more here, too
	end

	def jump(pos)
		if(reconnect_if_disconnected)
			@xc.playlist_set_next(pos).notifier do |res|
				@xc.playback_tickle # jump to the new position when it's set
			end

			return true
		else
			return false
		end
	end

	def jump_rel(pos)
		if(reconnect_if_disconnected)
			@xc.playlist_set_next_rel(pos).notifier do |res|
				@xc.playback_tickle # jump to the new position when it's set
			end

			return true
		else
			return false
		end
	end

	def get_pos_from_id(id)
		pos = []

		@playlist.each_index do |i|
			pos.push(i) if(@playlist[i][:server][:id] == id)
		end

		return pos
	end

	def close
		if(connected?)
			disconnect
		end
	end

	private

	def try_connect
		@xc = XmmsClient::XmmsClient.new(@client)

		begin
			@connected = true

			@xc.connect(ENV['XMMS_PATH'] || @ipcpath ||
				"unix:///tmp/xmms-ipc-#{ENV['USER']}")
			@xc.add_to_glib_mainloop
			call_callbacks(:connected)
		rescue XmmsClient::XmmsClient::XmmsClientError
			@connected = false
		end
		init_notifiers
		init_status
		init_pos
		init_id
		init_playlist
	end

	def reconnect_if_disconnected
		if(connected?)
			return true
		else
			try_connect

			return connected?
		end
	end

	def init_playlist
		# Get the initial playlist.
		@playlist = []
		@length = 0

		if(connected?)
			@xc.playlist_list.notifier do |res|
				list = res.value
				@length = list.length

				# Let's be nice and get metadata automatically for every song on
				# the playlist.
				if(@length == 0)
					call_callbacks(:playlist, {:type => CLR}, nil)
				else
					list.each {|id| append_entry(id)}
				end
			end
		end
	end

	def init_status
		# Set the initial playback status.
		@status = UNKNOWN

		if(connected?)
			@xc.playback_status.notifier(&method(:update_status))
		end
	end

	def init_pos
		# Get the initial playlist position.
		@pos = -1

		if(connected?)
			@xc.playlist_current_pos.notifier(&method(:update_pos))
		end
	end

	def init_id
		# Get the initial playback id.
		@id = 0

		if(connected?)
			@xc.playback_current_id.notifier(&method(:update_id))
		end
	end

	def insert_mediainfo(id, pos)
		@xc.medialib_get_info(id).notifier do |res|
			begin
				info = res.value
				info[:server][:url_decoded] = res.decode_url(info[:server][:url])

				@playlist[pos] = info
				call_callbacks(:playlist, {:type => ADD}, info)
			rescue
				call_callbacks(:playlist, {:type => UNKNOWN}, nil)
			end
		end
	end

	def append_entry(id)
		# Push on a fake entry while we wait for the mediainfo, just in case
		# this entry fails to resolve and is deleted before mediainfo is read.
		@playlist.push({:server => {:id => id}})
		insert_mediainfo(id, @length - 1)
	end

	def insert_entry(id, pos)
		# Push on a fake entry while we wait for the mediainfo, just in case
		# this entry fails to resolve and is deleted before mediainfo is read.
		@playlist.insert(pos, {:server => {:id => id}})
		insert_mediainfo(id, pos)
	end

	def update_playlist(res)
		info = res.value
		# Some of these broadcasts give us metadata with some added tags that
		# don't need to appear in the playlist.
		pos = info.delete(:position)
		newpos = info.delete(:newposition)

		# The actual playlist array is kept in sync here. The broadcasts are
		# propagated using callbacks.
		case(info.delete(:type))
			when(ADD)
				@length += 1
				append_entry(info[:id])
			when(INS)
				@length += 1
				insert_entry(info[:id], pos)
			when(MOV)
				@playlist.insert(newpos, @playlist.delete_at(pos))
				call_callbacks(:playlist, {:type => MOV, :old => pos,
					:new => newpos}, @playlist[newpos])
			when(REM)
				@length -= 1
				check = @playlist[pos][:server][:id]
				@playlist.delete_at(pos)
				if(check)
					call_callbacks(:playlist, {:type => REM, :pos => pos}, info)
				end
			when(CLR)
				@playlist = []
				@length = 0
				call_callbacks(:playlist, {:type => CLR}, info)
			when(SRT)
				init_playlist
				call_callbacks(:playlist, {:type => SRT}, info)
			when(SHF)
				init_playlist
				call_callbacks(:playlist, {:type => SHF}, info)
		end
	end

	def update_mediainfo(res)
		id = res.value

		@xc.medialib_get_info(id).notifier do |res|
			info = res.value
			info[:server][:url_decoded] = res.decode_url(info[:server][:url])
			positions = get_pos_from_id(info[:server][:id])

			positions.each do |pos|
				@playlist[pos] = info
			end
			call_callbacks(:medialib_entry, info)
		end
	end

	def update_pos(res)
		# When a song is playing that is not on the playlist, or when xmms2d has
		# just started and is not yet playing, the position will not be set. So,
		# it's set to -1 here if that is the case.
		@pos = begin
			res.value
		rescue XmmsClient::Result::ValueError
			UNKNOWN
		end

		call_callbacks(:current_pos, @pos)
	end

	def update_id(res)
		@id = res.value

		call_callbacks(:current_id, @id)
	end

	def update_status(res)
		@status = res.value

		call_callbacks(:playback_status, @status)
	end

	def update_config(res)
		config = res.value

		call_callbacks(:configval, {config[:name] => config[:value]})
	end

	def init_callbacks
		# Set up callbacks for the async code to keep us updated.
		@callbacks = {
			:connected => [],
			:disconnected => [],
			:configval => [],
			:current_id => [],
			:medialib_entry => [],
			:playback_status => [],
			:playlist => [],
			:current_pos => [],
			:playtime => []
		}
	end

	def init_notifiers
		if(connected?)
			@broadcasts = {
				:configval => @xc.broadcast_configval_changed(),
				:current_id => @xc.broadcast_playback_current_id(),
				:medialib_entry => @xc.broadcast_medialib_entry_changed(),
				:playback_status => @xc.broadcast_playback_status(),
				:playlist => @xc.broadcast_playlist_changed(),
				:current_pos => @xc.broadcast_playlist_current_pos()
			}
			@signals = {
			#	:playtime => @xc.signal_playback_playtime()
			}

			# This stuff is obvious, I need to add some more handler blocks, though.
			@broadcasts[:configval].notifier(&method(:update_config))
			@broadcasts[:current_id].notifier(&method(:update_id))
			@broadcasts[:medialib_entry].notifier(&method(:update_mediainfo))
			@broadcasts[:playback_status].notifier(&method(:update_status))
			@broadcasts[:playlist].notifier(&method(:update_playlist))
			@broadcasts[:current_pos].notifier(&method(:update_pos))

			@xc.on_disconnect(&method(:disconnect))
		end
	end

	def call_callbacks(key, *val)
		@callbacks[key].each {|cb| cb.call(*val)}
	end

	def disconnect
		@connected = false
		@status = UNKNOWN
		call_callbacks(:disconnected)

		@xc.remove_from_glib_mainloop
		[@broadcasts, @signals].each do |hash|
			hash.each_value do |res|
				res.disconnect
			end
		end
		@xc.delete!
	end
end
